/*
 * $Id: kedge.cc,v 1.6 2003/10/28 17:52:14 jpr5 Exp $
 *
 * kedge 1.3
 * Copyright (c) 2003 Jordan Ritter
 * Released into the Public Domain.
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>

#include <klocale.h>
#include <kaboutdata.h>
#include <kapplication.h>
#include <kcmdlineargs.h>
#include <kconfig.h>

#include <netwm_def.h>
#include <netwm.h>

#include <X11/Xlib.h>


static const char *version = "1.3";

static int (*old_XErrorHandler)(Display*, XErrorEvent *);
static unsigned int x_edge, y_edge;
static int max_desktop, dest_desktop;
static Display *display;
static NETRootInfo *r;
static Window root;
static int factor;


/*
 * dummy error handler to deal with the X11 server giving us bogus
 * CreateWindow events (on destroy), and other bogosities.
 */

int my_XErrorHandler(Display *d, XErrorEvent *e) {
    if (e->error_code == BadWindow)
        return 0;
    else return old_XErrorHandler(d, e);
}


/*
 * KDE_VERSION_MAJOR is used because versions 2.x <-> 3.x differ in
 * how the desktops are ordered..
 */

void DetectEdge(unsigned x, unsigned y) {
    if (x == 0) {

        r->activate();
#if KDE_VERSION_MAJOR == 2
        dest_desktop = r->currentDesktop() - factor;
#else
        dest_desktop = r->currentDesktop() - 1;
#endif

        /*
         * not sure if everything to the right of ``||'' will work
         * with KDE2, should test.
         */

        if (dest_desktop < 1 || dest_desktop == max_desktop/2)
            return;

        r->setCurrentDesktop(dest_desktop);
        XWarpPointer(display, None, root, 0, 0, 0, 0, x_edge-10, y);

    } else if (x == x_edge) {

        r->activate();
#if KDE_VERSION_MAJOR == 2
        dest_desktop = r->currentDesktop() + factor;
#else
        dest_desktop = r->currentDesktop() + 1;
#endif

        /*
         * not sure if everything to the right of ``||'' will work
         * with KDE2, should test.
         */

        if (dest_desktop > max_desktop || dest_desktop == max_desktop/2 + 1)
            return;

        r->setCurrentDesktop(dest_desktop);
        XWarpPointer(display, None, root, 0, 0, 0, 0, 10, y);

    } else if (factor > 1) {

        if (y == 0) {

            r->activate();
#if KDE_VERSION_MAJOR == 2
            dest_desktop = r->currentDesktop() - 1;
#else
            dest_desktop = r->currentDesktop() - max_desktop/factor;
#endif

            if (dest_desktop < 1)
                return;

            r->setCurrentDesktop(dest_desktop);
            XWarpPointer(display, None, root, 0, 0, 0, 0, x, y_edge-10);

        } else if (y == y_edge) {

            r->activate();
#if KDE_VERSION_MAJOR == 2
            dest_desktop = r->currentDesktop() + 1;
#else
            dest_desktop = r->currentDesktop() + max_desktop/factor;
#endif

            if (dest_desktop > max_desktop)
                return;

            r->setCurrentDesktop(dest_desktop);
            XWarpPointer(display, None, root, 0, 0, 0, 0, x, 10);

        }
    }
}


void setupKDE(Display *d, int argc, char **argv) {

    /*
     * First, let's get the current number of desktops.
     */

    r = new NETRootInfo(d, NET::NumberOfDesktops|
                        NET::DesktopGeometry|
                        NET::CurrentDesktop|
                        NET::VirtualRoots);
    r->activate();
    max_desktop = r->numberOfDesktops();

    /*
     * Below is a little bit of fun.  In order to access kicker's
     * config to figure out the current panel size, we need a
     * KApplication around (we really need a KInstance around, but I
     * couldn't figure out how to instantiate one properly).  So, once
     * we have a KApplication, we won't segfault when we try to
     * instantiate KConfig (which calls KInstance::dirs()) and reads
     * the config setting we're after.
     */

    KAboutData aboutData( "kedge", I18N_NOOP("Edge-sensitive desktop switcher."),
                          version,  I18N_NOOP("Edge-sensitive desktop switcher."),
                          KAboutData::License_Artistic, "(c) 2003 Jordan Ritter");
    KCmdLineArgs::init(argc, argv, &aboutData );
    KApplication a;
    KConfig c("kickerrc", true, false);
    c.setGroup("General");

    /*
     * 0 = tiny
     * 1 = small
     * 2 = normal
     * 3 = large
     * 4 = custom
     */

    switch (c.readUnsignedNumEntry("Size", 0)) {
        case 0: // tiny, small
        case 1: {
            factor = 1;
        } break;

        case 2: // normal, large, custom
        case 3:
        case 4: {
            factor = 2;
        } break;
    }
}


void spy(Window r) {
    Window p, *children;
    unsigned nchildren;

    XWindowAttributes a = {0,};
    XGetWindowAttributes(display, r, &a);
    XSelectInput(display, r, a.your_event_mask|PointerMotionMask);

    if (!XQueryTree(display, r, &r, &p, &children, &nchildren)) {
        fprintf(stderr, "warning: couldn't query window tree\n");
        return;
    }

    if (!nchildren)
        return;

    for (unsigned i = 0; i < nchildren; i++)
        spy(children[i]);

    XFree(children);
}


int main(int argc, char **argv) {

    old_XErrorHandler = XSetErrorHandler(my_XErrorHandler);

    if (!(display = XOpenDisplay(":0.0"))) {
        fprintf(stderr, "fatal: couldn't open display\n");
        return 1;
    }

    root = DefaultRootWindow(display);

    setupKDE(display, argc, argv);

    /*
     * find out how big the screen is
     */

    XWindowAttributes a = {0,};
    XGetWindowAttributes(display, root, &a);

    x_edge = a.width-1;
    y_edge = a.height-1;

    /*
     * notify us of changes to any children from the root window.
     */

    XSelectInput(display, root, a.your_event_mask|SubstructureNotifyMask);

    /*
     * And now spy on all child windows.
     */

    spy(root);

    while (1) {
        XEvent e;
        XNextEvent(display, &e);

        switch (e.type) {
            /*
             * Snag events from new windows.
             */

            case CreateNotify: {
                XCreateWindowEvent *ce = (XCreateWindowEvent*)&e;
                XWindowAttributes a = {0,};
                XGetWindowAttributes(display, ce->window, &a);
                XSelectInput(display, ce->window, a.your_event_mask|PointerMotionMask);
            } break;

            /*
             * And when we move, look for an edge.
             */

            case MotionNotify: {
                XMotionEvent *me = (XMotionEvent*)&e;
                DetectEdge(me->x_root, me->y_root);
            } break;
        }
    }

    return 0;
}

